"""Utilities for working with source language lexicons.

Resources include:
- [UBS Dictionary of the Greek New
  Testament](https://github.com/ubsicap/ubs-open-license/tree/main/dictionaries/greek)
- [UBS Dictionary of Biblical
  Hebrew](https://github.com/ubsicap/ubs-open-license/tree/main/dictionaries/hebrew)

"""

from .LexiconEntry import LEXSense, LEXMeaning, BaseForm, Lexicon_Entry
from .ubsdgnt import UBSParser

__all__ = [
    # LexiconEntry
    "LEXSense",
    "LEXMeaning",
    "BaseForm",
    "Lexicon_Entry",
    # ubsdgnt
    "UBSParser",
]



"""Set up paths for loaders."""

# haven't figured out Django loading yet
# from django.conf import settings

# LOADER_DATA_PATH = settings.BASE_DIR / "loader_data"

from pathlib import Path
from collections import UserDict
from csv import DictReader

from pydantic import BaseModel  # , ValidationError, conint, constr


SERVER_ROOT = Path(__file__).parent.parent.parent

LOADER_DATA_PATH = SERVER_ROOT / "loader_data"
LOADER_SOURCE_PATH = SERVER_ROOT / "alignment_api/pipelines"


class Reader(UserDict):
    """Read external data and marshall for loading into Django.

    This is generalized for use where the source data is in TSV, with
    column headers that match a BaseModel instance, and a designated
    identity element.

    """

    # subs must define these
    tsvpath: Path = None
    idattr: str = ""
    model: BaseModel = None

    def __init__(self) -> None:
        """Initialize a Reader instance."""
        super().__init__()
        assert self.tsvpath, "tsvpath must be defined"
        assert self.idattr, "idattr must be defined"
        assert self.model, "model must be defined"
        with self.tsvpath.open() as f:
            reader = DictReader(f, delimiter="\t")
            self.data = {
                idattr: self.model(**row) for row in reader if (idattr := row[self.idattr])
            }



"""Read the UBS dictionaries into lookup dicts.

The dataclass reflect the XML hierarchy (selectively):
Lexicon_Entry.baseforms ->
BaseForm.lexMeanings ->
LEXMeaning.lexSenses ->
LEXSense
"""

from dataclasses import dataclass, field
from typing import Any, Optional


@dataclass
class LEXSense:
    """Represents the same element from UBSGNT as a Python dataclass."""

    # no distinct identifier: 1-1 with LEXMeaning for Greek
    # the language of the definition
    languageCode: str
    glosses: list[str]
    definitionShort: str
    definitionLong: str
    comments: list[str]
    # link upward: beware recursion
    parent: Optional["LEXMeaning"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<LEXSense: {';'.join(self.glosses)}>"

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        fields = ["languageCode", "glosses", "definitionShort", "definitionLong", "comments"]
        return {f: getattr(self, f) for f in fields}


@dataclass
class LEXMeaning:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    isBiblicalTerm: str
    entryCode: str
    lexDomains: list[str]
    lexSubDomains: list[str]
    # appears to only be one LEXSense per LEXMeaning? But the
    # structure allows for more
    lexSenses: list[LEXSense]
    # using MARBLE-style references
    # this includes the deuterocanon, which seems to often have
    # multiple meanings for a single reference
    # some unusual ones like '04500900300002{N:002}'
    lexReferences: list[str]
    # link upward: beware recursion
    parent: Optional["BaseForm"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<LEXMeaning: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        fields = ["identifier", "isBiblicalTerm", "entryCode", "lexDomains", "lexSubDomains"]
        return {f: getattr(self, f) for f in fields}


@dataclass
class BaseForm:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    partsofspeech: list[str]
    lexMeanings: list[LEXMeaning]
    # FUTURE: incorporate RelatedLemmas.L
    # link upward: beware recursion
    parent: Optional["Lexicon_Entry"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<BaseForm: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        fields = ["identifier", "partsofspeech"]
        return {f: getattr(self, f) for f in fields}


@dataclass
class Lexicon_Entry:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    lemma: str
    strongCodes: list[str]
    # FUTURE: capture Notes (see Lexicon_Entry Id="003076000000000")
    # multiple values if different parts of speech,
    # e.g. <Lexicon_Entry: 000266000000000>
    baseforms: list[BaseForm]

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<Lexicon_Entry: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def get_meanings(self) -> list[LEXMeaning]:
        """Return a list of of LEXMeanings for this entry."""
        return [lm for bf in self.baseforms for lm in bf.lexMeanings]



"""Load external license data.

Not much validation required since this is a simple input file.
"""


from pydantic import BaseModel  # , ValidationError, conint, constr

from . import config


class License(BaseModel):
    """Manage data that is read in from file on a license.

    This should stay in sync with alignment_api.models.License.
    """

    license_id: str
    license_url: str
    attribution: bool
    sharealike: bool
    noderivs: bool
    noncommercial: bool


class Reader(config.Reader):
    """Read external data and marshall for loading into Django."""

    tsvpath = config.LOADER_DATA_PATH / "license.tsv"
    idattr = "license_id"
    model = License



"""Parse UBSGNT into dataclasses.

>>> from acai.lexicon import UBSParser
>>> ubsprs = UBSParser()
>>> len(ubsprs)

"""

from collections import UserDict, defaultdict
from typing import Any
import unicodedata

import requests
from lxml import etree

from biblelib.word import bcvwpid

from .LexiconEntry import LEXSense, LEXMeaning, BaseForm, Lexicon_Entry


def get_ubsdgnt(langcode: str = "eng") -> str:
    """Return a string for the readable content from Github."""
    pathprefix: str = (
        "https://github.com/ubsicap/ubs-open-license/raw/main/dictionaries/greek/XML/UBSGreekNTDic-v1.0"
    )
    langcodes = {
        "eng": "en",
        "fra": "fr",
        "spa": "sp",
        # not ISO 639-3
        "zhs": "zhs",
        "zht": "zht",
    }
    return f"{pathprefix}-{langcodes[langcode]}.XML"


class UBSParser(UserDict):
    """Manage data read from UBSGNT."""

    # for lookup methods
    lemma_entries: dict[str, Lexicon_Entry] = {}
    entryCode_meanings: dict[str, LEXMeaning] = {}
    # sometimes multiple meanings for a word-level reference
    reference_meanings: dict[str, list[LEXMeaning]] = defaultdict(list)

    def __init__(self, langcode: str = "eng"):
        """Initialize the parser."""
        super().__init__()
        # these store elements for debugging
        # unfollowed import problem, so values are Any
        self.lexicon_entries: dict[str, Any] = {}
        self.baseform_elements: dict[str, Any] = {}
        self.lexmeaning_elements: dict[str, Any] = {}
        self.dicturi = get_ubsdgnt(langcode)
        r = requests.get(self.dicturi)
        assert r.status_code == 200, f"Failed to get content from {self.dicturi}"
        # normalize everything so matching works
        self.root = etree.fromstring(r.text)
        # iterate over Lexicon_Entry elements
        # for Greek, each Lexicon_Entry.lemma is unique, so could use
        # lemma for keys rather than identifier. Not sure about Hebrew though
        self.data = {
            thislexe.identifier: thislexe
            for lexentry in self.root.xpath("//Lexicon_Entry")
            if (thislexe := self.do_lexicon_entry(lexentry))
        }

    @staticmethod
    def _first_if(xpathexpr: str) -> str:
        """Return first string value in the xpath result, or empty string."""
        return xpathexpr[0] if len(xpathexpr) > 0 else ""

    # type complaint with this
    #    def do_senses(self, lexmean: etree._Element) -> list[LEXSense]:
    def do_senses(self, lexmean: Any) -> list[LEXSense]:
        """Process LEXSense content.

        Returns a list with a single LEXSense instance. This leaves
        room for other data that might have multiple.

        """
        # only ever one LEXSense for Greek
        lexsense = lexmean.xpath("LEXSenses/LEXSense")[0]
        lsdict = dict(
            languageCode=lexsense.xpath("@LanguageCode")[0],
            glosses=lexsense.xpath("Glosses/Gloss/text()"),
            definitionShort=self._first_if(lexsense.xpath("DefinitionShort/text()")),
            definitionLong=self._first_if(lexsense.xpath("DefinitionLong/text()")),
            comments=lexsense.xpath("Comments/Comment/text()"),
        )
        return [LEXSense(**lsdict)]

    # def do_meaning(self, lexmean: etree._Element) -> LEXMeaning:
    def do_meaning(self, lexmean: Any) -> LEXMeaning:
        """Process LEXMeaning contents."""
        # stash for debugging
        lexmean_id: str = lexmean.xpath("@Id")[0]
        self.lexmeaning_elements[lexmean_id] = lexmean
        try:
            lexsenses = self.do_senses(lexmean)
        except Exception as e:
            print(f"Failed on {lexmean}, {lexmean_id}\n{e}")
            lexsenses = []
        try:
            lexReferences = [
                bcvwpid.fromubs(ref).get_id()
                for ref in lexmean.xpath("LEXReferences/LEXReference/text()")
            ]
        except Exception as e:
            print(f"Failed on references for {lexmean}, {lexmean_id}\n{e}")
            lexReferences = []
        lexmeaning = LEXMeaning(
            identifier=lexmean_id,
            isBiblicalTerm=self._first_if(lexmean.xpath("@IsBiblicalTerm")),
            entryCode=self._first_if(lexmean.xpath("@EntryCode")),
            # does this capture multiple values correctly?
            lexDomains=lexmean.xpath("LEXDomains/LEXDomain/text()"),
            lexSubDomains=lexmean.xpath("LEXSubDomains/LEXSubDomain/text()"),
            lexSenses=lexsenses,
            lexReferences=lexReferences,
        )
        for lexs in lexsenses:
            lexs.parent = lexmeaning
        return lexmeaning

    def do_baseform(self, baseformel: Any) -> BaseForm:
        """Process BaseForm contents."""
        # stash for debugging
        baseform_id: str = baseformel.xpath("@Id")[0]
        self.baseform_elements[baseform_id] = baseformel
        baseform = BaseForm(
            identifier=baseform_id,
            partsofspeech=baseformel.xpath("PartsOfSpeech/PartOfSpeech/text()"),
            lexMeanings=[
                self.do_meaning(lm)
                for lm in baseformel.xpath("LEXMeanings/LEXMeaning")
                if self._first_if(lm.xpath("@IsBiblicalTerm")) != "N"
            ],
        )
        for lm in baseform.lexMeanings:
            lm.parent = baseform
        return baseform

    def do_lexicon_entry(self, lexentry: Any) -> Lexicon_Entry:
        """Process Lexicon_Entry contents."""
        lexentry_id: str = lexentry.xpath("@Id")[0]
        self.lexicon_entries[lexentry_id] = lexentry
        # PROCESS do_baseform here
        thislexe = Lexicon_Entry(
            identifier=lexentry_id,
            lemma=unicodedata.normalize("NFKC", lexentry.xpath("@Lemma")[0]),
            strongCodes=lexentry.xpath("StrongCodes/Strong/text()"),
            # this assumes a single BaseForm per lexentry
            baseforms=[self.do_baseform(bf) for bf in lexentry.xpath("BaseForms/BaseForm")],
        )
        for bf in thislexe.baseforms:
            bf.parent = thislexe
        return thislexe

    # accessors
    # only build these on demand

    # just make lemma the key for self.data?
    def get_lemma_entries(self, lemma: str) -> Lexicon_Entry:
        """Return the Lexicon_Entry instance for lemma."""
        if not self.lemma_entries:
            self.lemma_entries = {entry.lemma: entry for entry in self.data.values()}
        return self.lemma_entries[lemma]

    def get_entryCode_meanings(self, entryCode: str) -> LEXMeaning:
        """Return the LEXMeaning instance for an entryCode."""
        if not self.entryCode_meanings:
            self.entryCode_meanings = {
                meaning.entryCode: meaning
                for entry in self.data.values()
                for baseform in entry.baseforms
                for meaning in baseform.lexMeanings
            }
        return self.entryCode_meanings[entryCode]

    def get_reference_meanings(self, reference: str) -> list[LEXMeaning]:
        """Return the LEXMeaning instances for an reference."""
        if not self.reference_meanings:
            for entry in self.data.values():
                for baseform in entry.baseforms:
                    for meaning in baseform.lexMeanings:
                        for ref in meaning.lexReferences:
                            self.reference_meanings[ref].append(meaning)
        return self.reference_meanings[reference]

    def get_strongs_lemmas(self) -> dict[str, list[Lexicon_Entry]]:
        """Return a mapping from strongs to lemmas."""
        strongs_lemmas: dict[str, list[Lexicon_Entry]] = defaultdict(list)
        for entry in self.data.values():
            for strong in entry.strongCodes:
                strongs_lemmas[strong].append(entry)
        return strongs_lemmas

    def get_strongs_meanings(self) -> dict[str, list[LEXMeaning]]:
        """Return a mapping from strongs to lexMeanings."""
        strongs_lemmas: dict[str, list[LEXMeaning]] = defaultdict(list)
        for entry in self.data.values():
            for baseform in entry.baseforms:
                for meaning in baseform.lexMeanings:
                    for strong in entry.strongCodes:
                        strongs_lemmas[strong].append(meaning)
        return strongs_lemmas

